# -*- coding: utf-8 -*-
"""
hyper/tls
~~~~~~~~~

Contains the TLS/SSL logic for use in hyper.
"""
import os.path as path
from .common.exceptions import MissingCertFile
from .compat import ignore_missing, ssl


NPN_PROTOCOL = 'h2'
H2_NPN_PROTOCOLS = [NPN_PROTOCOL, 'h2-16', 'h2-15', 'h2-14']
SUPPORTED_NPN_PROTOCOLS = H2_NPN_PROTOCOLS + ['http/1.1']

H2C_PROTOCOL = 'h2c'
VERIFY_CERT = False

# We have a singleton SSLContext object. There's no reason to be creating one
# per connection.
_context = None

# Work out where our certificates are.
cert_loc = path.join(path.dirname(__file__), 'certs.pem')


def wrap_socket(sock, server_hostname, ssl_context=None, force_proto=None):
    """
    A vastly simplified SSL wrapping function. We'll probably extend this to
    do more things later.
    """

    global _context

    if ssl_context:
        # if an SSLContext is provided then use it instead of default context
        _ssl_context = ssl_context
    else:
        # create the singleton SSLContext we use
        if _context is None:  # pragma: no cover
            _context = init_context()
        _ssl_context = _context

    # the spec requires SNI support
    ssl_sock = _ssl_context.wrap_socket(sock, server_hostname=server_hostname)
    # Setting SSLContext.check_hostname to True only verifies that the
    # post-handshake servername matches that of the certificate. We also need
    # to check that it matches the requested one.
    if _ssl_context.check_hostname:  # pragma: no cover
        try:
            ssl.match_hostname(ssl_sock.getpeercert(), server_hostname)
        except AttributeError:
            ssl.verify_hostname(ssl_sock, server_hostname)  # pyopenssl

    # Allow for the protocol to be forced externally.
    proto = force_proto

    # ALPN is newer, so we prefer it over NPN. The odds of us getting
    # different answers is pretty low, but let's be sure.
    with ignore_missing():
        if proto is None:
            proto = ssl_sock.selected_alpn_protocol()

    with ignore_missing():
        if proto is None:
            proto = ssl_sock.selected_npn_protocol()

    return (ssl_sock, proto)


def init_context(cert_path=None, cert=None, cert_password=None):
    """
    Create a new ``SSLContext`` that is correctly set up for an HTTP/2
    connection. This SSL context object can be customized and passed as a
    parameter to the :class:`HTTPConnection <hyper.HTTPConnection>` class.
    Provide your own certificate file in case you don’t want to use hyper’s
    default certificate. The path to the certificate can be absolute or
    relative to your working directory.

    :param cert_path: (optional) The path to the certificate file of
        “certification authority” (CA) certificates
    :param cert: (optional) if string, path to ssl client cert file (.pem).
        If tuple, ('cert', 'key') pair.
        The certfile string must be the path to a single file in PEM format
        containing the certificate as well as any number of CA certificates
        needed to establish the certificate’s authenticity. The keyfile string,
        if present, must point to a file containing the private key in.
        Otherwise the private key will be taken from certfile as well.
    :param cert_password: (optional) The password argument may be a function to
        call to get the password for decrypting the private key. It will only
        be called if the private key is encrypted and a password is necessary.
        It will be called with no arguments, and it should return a string,
        bytes, or bytearray. If the return value is a string it will be
        encoded as UTF-8 before using it to decrypt the key. Alternatively a
        string, bytes, or bytearray value may be supplied directly as the
        password argument. It will be ignored if the private key is not
        encrypted and no password is needed.
    :returns: An ``SSLContext`` correctly set up for HTTP/2.
    """
    cafile = cert_path or cert_loc
    if not cafile or not path.exists(cafile):
        err_msg = ("No certificate found at " + str(cafile) + ". Either " +
                   "ensure the default cert.pem file is included in the " +
                   "distribution or provide a custom certificate when " +
                   "creating the connection.")
        raise MissingCertFile(err_msg)

    context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
    context.set_default_verify_paths()
    context.load_verify_locations(cafile=cafile)

    if VERIFY_CERT:
        context.verify_mode = ssl.CERT_REQUIRED
        context.check_hostname = True

    with ignore_missing():
        context.set_npn_protocols(SUPPORTED_NPN_PROTOCOLS)

    with ignore_missing():
        context.set_alpn_protocols(SUPPORTED_NPN_PROTOCOLS)

    # required by the spec
    context.options |= ssl.OP_NO_COMPRESSION

    if cert is not None:
        try:
            basestring
        except NameError:
            basestring = (str, bytes)
        if not isinstance(cert, basestring):
            context.load_cert_chain(cert[0], cert[1], cert_password)
        else:
            context.load_cert_chain(cert, password=cert_password)

    return context





import sys
cache = {}
ocache = {}
all = []

try:
	from google.protobuf import descriptor_pb2

except ImportError:
	from ...uninstalled import Uninstalled
	find_object = discover = Uninstalled ('protobuf')

else:
	def discover ():
		global cache
		cache = {}
		for k, v in sys.modules.items ():
			if k.endswith ("_pb2") and not k.startswith ("google."):
				desriptor = v.DESCRIPTOR
				proto = descriptor_pb2.FileDescriptorProto.FromString(desriptor.serialized_pb)
				for service in proto.service:
					for method in service.method:
						cache ["%s.%s/%s" % (desriptor.package, service.name, method.name)] = (
							v,
							(method.input_type.split (".")[-1], method.client_streaming),
							(method.output_type.split (".")[-1], method.server_streaming),
							method.options
						)
				all.append (v)
	discover ()

	def find_object (module, name_stream):
		global all, ocache

		name, isstream = name_stream
		if name in ocache:
			return ocache [name]

		res = None
		if hasattr (module, name):
			res = getattr (module, name), isstream
		else:
			for each in all:
				if hasattr (each, name):
					res = getattr (each, name), isstream
					break

		if res:
			ocache [name] = res

		return res

def find_type (uri):
	global cache, ocache
	if not ocache: discover ()
	return find_input (uri), find_output (uri)

def find_output (uri):
	global cache, ocache
	if not ocache: discover ()
	try:
		module, it, ot, opt = cache.get (uri)
	except TypeError:
		raise KeyError
	return find_object (module, ot)

def find_input (uri):
	global cache, ocache
	if not ocache: discover ()
	try:
		module, it, ot, opt = cache.get (uri)
	except TypeError:
		raise KeyError
	return find_object (module, it)





# -*- coding: utf-8 -*-
class RFC3986Exception(Exception):
    pass


class InvalidAuthority(RFC3986Exception):
    def __init__(self, authority):
        super(InvalidAuthority, self).__init__(
            "The authority ({0}) is not valid.".format(authority))


class InvalidPort(RFC3986Exception):
    def __init__(self, port):
        super(InvalidPort, self).__init__(
            'The port ("{0}") is not valid.'.format(port))


class ResolutionError(RFC3986Exception):
    def __init__(self, uri):
        super(ResolutionError, self).__init__(
            "{0} is not an absolute URI.".format(uri.unsplit()))



# -*- coding: utf-8 -*-
"""
hyper
~~~~~~

A module for providing an abstraction layer over the differences between
HTTP/1.1 and HTTP/2.
"""
import logging

from .common.connection import HTTPConnection
from .http20.connection import HTTP20Connection
from .http20.response import HTTP20Response, HTTP20Push
from .http11.connection import HTTP11Connection
from .http11.response import HTTP11Response

# Throw import errors on Python <2.7 and 3.0-3.2.
import sys as _sys
if _sys.version_info < (2, 7) or (3, 0) <= _sys.version_info < (3, 3):
    raise ImportError(
        "hyper only supports Python 2.7 and Python 3.3 or higher."
    )

__all__ = [
    HTTPConnection,
    HTTP20Response,
    HTTP20Push,
    HTTP20Connection,
    HTTP11Connection,
    HTTP11Response,
]

# Set default logging handler.
logging.getLogger(__name__).addHandler(logging.NullHandler())

__version__ = '0.7.0'

def enable_verify_cert (flag):
    from . import tls
    tls.VERIFY_CERT = flag


from aioquic.h3.events import H3Event, DataReceived, HeadersReceived, PushPromiseReceived
from dataclasses import dataclass

@dataclass
class ConnectionShutdownInitiated (H3Event):
    last_stream_id: int

@dataclass
class PushCanceled (H3Event):
    push_id: int

@dataclass
class MaxPushIdReceived (H3Event):
    push_id: int

@dataclass
class DuplicatePushReceived (H3Event):
    stream_id: int
    push_id: int
    stream_ended: bool = True



from aioquic.h3.connection import H3Connection
from aioquic.h3.connection import encode_frame, encode_uint_var
try:
    from aioquic.h3.connection import parse_uint_var
except ImportError:
    from aioquic.h3.connection import parse_max_push_id as parse_uint_var
from aioquic.h3.connection import FrameType, StreamType, FrameUnexpected
from aioquic.h3.exceptions import NoAvailablePushIDError
from .events import PushCanceled, MaxPushIdReceived, ConnectionShutdownInitiated, DuplicatePushReceived
from rs4.annotations import override

class H3Connection (H3Connection):
    def __init__ (self, quic):
        super ().__init__ (quic)
        self._max_client_bidi_stream_id = 0 if not self._is_client else None
        self._push_map = {}
        self._canceled_push_ids = set ()
        self._uncaught_events = []

    @override
    def send_push_promise (self, stream_id, headers):
        push_stream_id = super ().send_push_promise (stream_id, headers)
        self._push_map [push_stream_id] = self._next_push_id - 1
        return push_stream_id

    def get_push_id (self, stream_id):
        try:
            return self._push_map [stream_id]
        except KeyError:
            raise AssertionError("No such stream ID")

    def cancel_push (self, push_id):
        self._canceled_push_ids.add (push_id)
        self._quic.send_stream_data (self._local_control_stream_id, encode_frame(FrameType.CANCEL_PUSH, encode_uint_var(push_id)))

    def send_duplicate_push (self, stream_id, push_id):
        assert not self._is_client, "Only servers may send a duplicate push."
        assert push_id < self._max_push_id, "Given push ID is never sent"
        assert (
            push_id not in self._canceled_push_ids
        ), "Given push ID is canceled by client"

        self._quic.send_stream_data (stream_id, encode_frame(FrameType.DUPLICATE_PUSH, encode_uint_var(push_id)), True)

    def shutdown (self, last_stream_id = None):
        assert not self._is_client, "Client must not send a goaway frame"
        if last_stream_id is None:
            last_stream_id = self._max_client_bidi_stream_id

        else:
            if last_stream_id == -1:
                last_stream_id = -4
            assert last_stream_id % 4 == 0 and (-4 <= last_stream_id <= self._max_client_bidi_stream_id), "Unissued request stream"

        frame = encode_frame(FrameType.GOAWAY, encode_uint_var(last_stream_id + 4))
        self._quic.send_stream_data (self._local_control_stream_id, frame)

    @override
    def handle_event (self, event):
        http_events = super ().handle_event (event)
        _uncaught_events, self._uncaught_events = self._uncaught_events, []
        return http_events + _uncaught_events

    @override
    def _handle_control_frame (self, frame_type, frame_data):
        http_events = super ()._handle_control_frame (frame_type, frame_data)
        assert http_events is None, '_handle_control_frame() returns something, update this method'

        if frame_type == FrameType.MAX_PUSH_ID:
            self._uncaught_events.append (MaxPushIdReceived (push_id = self._max_push_id))

        elif frame_type == FrameType.CANCEL_PUSH:
            _push_id = parse_uint_var (frame_data)
            self._canceled_push_ids.add (_push_id)
            self._uncaught_events.append (PushCanceled (push_id = _push_id))

        elif frame_type == FrameType.GOAWAY:
            _last_stream_id = max (-1, parse_uint_var (frame_data) - 4)
            self._uncaught_events.append (ConnectionShutdownInitiated (last_stream_id = _last_stream_id))

    @override
    def _handle_request_or_push_frame (self, frame_type, frame_data, stream, stream_ended):
        if frame_type == FrameType.HEADERS:
            if not self._is_client and stream.push_id is None:
                self._max_client_bidi_stream_id = max (stream.stream_id, self._max_client_bidi_stream_id)

        try:
            events = super ()._handle_request_or_push_frame (frame_type, frame_data, stream, stream_ended)
        except FrameUnexpected:
            if frame_type != FrameType.DUPLICATE_PUSH:
                raise
            _push_id = parse_uint_var (frame_data)
            events = [DuplicatePushReceived (stream_id = stream.stream_id, push_id = _push_id)]
        return events



from rs4 import asyncore
import re
import time
import sys
import threading
from ..client import baseconnect

DEBUG = False

class OperationalError (Exception): pass
class SQLError (Exception): pass

class ConnectProxy (baseconnect.ConnectProxy):
    def execute_late (self):
        if self._canceled (): # MUST BE CALLED for deactivating asyncon
            return
        request, rs, handle_request_failed = self._args
        self._asyncon.set_timeout (self._timeout)
        try:
            self._asyncon.execute (request)
        except:
            handle_request_failed (rs, self._asyncon)
        self._args = None


class DBConnect (baseconnect.BaseConnect):
    def __init__ (self, address, params = None, lock = None, logger = None):
        super ().__init__ (lock, logger)
        self.address = address
        self.params = params

        # need if there's not any request yet
        self.execute_count = 0
        self.request = None
        self.has_result = False

        self.set_auth ()
        self.set_event_time ()

    def set_auth (self):
        self.dbname, self.user, self.password = "", "", ""
        if not self.params:
            return
        self.dbname, auth = self.params
        if not auth:
            return
        if type (auth) is not tuple:
            self.user = auth
        else:
            try: self.user, self.password = auth
            except ValueError: self.username = auth

    def close_if_over_keep_live (self):
        if self.connected and time.time () - self.event_time > self.keep_alive:
            self.disconnect ()

    def duplicate (self):
        new_asyncon = self.__class__ (self.address, self.params, self.lock, self.logger)
        new_asyncon.keep_alive = self.keep_alive
        new_asyncon.backend = self.backend
        return new_asyncon

    def get_proto (self):
        # call by culster_manager
        return None

    def handle_abort (self):
        # call by dist_call
        self.close ()

    def close (self, deactive = 1):
        addr = type (self.address) is tuple and ("%s:%d" % self.address) or str (self.address)
        if deactive:
            self.logger ("[info] ..dbo %s has been closed" % addr)
            self.set_active (False)
            self.request = None

    def clean_shutdown_control (self, phase, time_in_this_phase):
        self._no_more_request = True
        if self.isactive ():
            return 1
        else:
            self.close ()
            self._no_more_request = False
            return 0

    def maintern (self, object_timeout):
        # when in map, mainteren by lifetime with zombie_timeout
        if self.is_channel_in_map ():
            return False
        idle = time.time () - self.event_time
        if idle > object_timeout:
            self.close (1)
            return True # deletable
        return False

    def reconnect (self):
        self.disconnect ()
        self.connect ()

    def disconnect (self):
        # keep request, just close temporary
        self.close (deactive = 0)

    def close_case (self):
        raise NotImplementedError

    def isconnected (self):
        # self.connected should be defined at __init__ or asyncore
        return self.connected

    def get_execute_count (self):
        return self.execute_count

    def connect (self, force = 0):
        raise NotImplementedError("must be implemented in subclass")

    def handle_timeout (self):
        self.handle_close (OperationalError ("Operation Timeout"))

    def handle_error (self):
        dummy, t, v, info = asyncore.compact_traceback ()
        self.has_result = False
        self.logger.trace ()
        self.handle_close (v)

    def handle_close (self, expt = None):
        if self.expt is None:
            self.expt = expt
        self.close_case_with_end_tran ()
        self.close ()

    def close_case_with_end_tran (self):
        self.end_tran ()
        self.close_case ()

    #-----------------------------------------------------
    # DB methods
    #-----------------------------------------------------
    def fetchall (self):
        raise NotImplementedError

    def end_tran (self):
        raise NotImplementedError

    def begin_tran (self, request):
        if self._no_more_request:
            raise OperationalError ("Entered Shutdown Process")
        self.request = request
        self._history = []
        self.out_buffer = ''
        self.has_result = False
        self.expt = None
        self.execute_count += 1
        self.close_if_over_keep_live ()
        self.set_event_time ()

    def execute (self, request, *args, **kargs):
        self.begin_tran (request)
        raise NotImplementedError("must be implemented in subclass")


class AsynDBConnect (DBConnect):
    def is_channel_in_map (self, map = None):
        if map is None:
            map = self._map
        return self._fileno in map

    def del_channel (self, map=None):
        # do not remove self._fileno
        fd = self._fileno
        if map is None:
            map = self._map
        if fd in map:
            del map[fd]




# -*- coding: utf-8 -*-
# Copyright (c) 2014 Rackspace
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
rfc3986.misc
~~~~~~~~~~~~

This module contains important constants, patterns, and compiled regular
expressions for parsing and validating URIs and their components.
"""

import re

# These are enumerated for the named tuple used as a superclass of
# URIReference
URI_COMPONENTS = ['scheme', 'authority', 'path', 'query', 'fragment']

important_characters = {
    'generic_delimiters': ":/?#[]@",
    'sub_delimiters': "!$&'()*+,;=",
    # We need to escape the '*' in this case
    're_sub_delimiters': "!$&'()\*+,;=",
    'unreserved_chars': ('ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'
                         '0123456789._~-'),
    # We need to escape the '-' in this case:
    're_unreserved': 'A-Za-z0-9._~\-',
    }
# For details about delimiters and reserved characters, see:
# http://tools.ietf.org/html/rfc3986#section-2.2
GENERIC_DELIMITERS = set(important_characters['generic_delimiters'])
SUB_DELIMITERS = set(important_characters['sub_delimiters'])
RESERVED_CHARS = GENERIC_DELIMITERS.union(SUB_DELIMITERS)
# For details about unreserved characters, see:
# http://tools.ietf.org/html/rfc3986#section-2.3
UNRESERVED_CHARS = set(important_characters['unreserved_chars'])
NON_PCT_ENCODED = RESERVED_CHARS.union(UNRESERVED_CHARS).union('%')

# Extracted from http://tools.ietf.org/html/rfc3986#appendix-B
component_pattern_dict = {
    'scheme': '[^:/?#]+',
    'authority': '[^/?#]*',
    'path': '[^?#]*',
    'query': '[^#]*',
    'fragment': '.*',
    }

# See http://tools.ietf.org/html/rfc3986#appendix-B
# In this case, we name each of the important matches so we can use
# SRE_Match#groupdict to parse the values out if we so choose. This is also
# modified to ignore other matches that are not important to the parsing of
# the reference so we can also simply use SRE_Match#groups.
expression = ('(?:(?P<scheme>{scheme}):)?(?://(?P<authority>{authority}))?'
              '(?P<path>{path})(?:\?(?P<query>{query}))?'
              '(?:#(?P<fragment>{fragment}))?'
              ).format(**component_pattern_dict)

URI_MATCHER = re.compile(expression)

# #########################
# Authority Matcher Section
# #########################

# Host patterns, see: http://tools.ietf.org/html/rfc3986#section-3.2.2
# The pattern for a regular name, e.g.,  www.google.com, api.github.com
reg_name = '(({0})*|[{1}]*)'.format(
    '%[0-9A-Fa-f]{2}',
    important_characters['re_sub_delimiters'] +
    important_characters['re_unreserved']
    )
# The pattern for an IPv4 address, e.g., 192.168.255.255, 127.0.0.1,
ipv4 = '(\d{1,3}.){3}\d{1,3}'
# Hexadecimal characters used in each piece of an IPv6 address
hexdig = '[0-9A-Fa-f]{1,4}'
# Least-significant 32 bits of an IPv6 address
ls32 = '({hex}:{hex}|{ipv4})'.format(hex=hexdig, ipv4=ipv4)
# Substitutions into the following patterns for IPv6 patterns defined
# http://tools.ietf.org/html/rfc3986#page-20
subs = {'hex': hexdig, 'ls32': ls32}

# Below: h16 = hexdig, see: https://tools.ietf.org/html/rfc5234 for details
# about ABNF (Augmented Backus-Naur Form) use in the comments
variations = [
    #                            6( h16 ":" ) ls32
    '(%(hex)s:){6}%(ls32)s' % subs,
    #                       "::" 5( h16 ":" ) ls32
    '::(%(hex)s:){5}%(ls32)s' % subs,
    # [               h16 ] "::" 4( h16 ":" ) ls32
    '(%(hex)s)?::(%(hex)s:){4}%(ls32)s' % subs,
    # [ *1( h16 ":" ) h16 ] "::" 3( h16 ":" ) ls32
    '((%(hex)s:)?%(hex)s)?::(%(hex)s:){3}%(ls32)s' % subs,
    # [ *2( h16 ":" ) h16 ] "::" 2( h16 ":" ) ls32
    '((%(hex)s:){0,2}%(hex)s)?::(%(hex)s:){2}%(ls32)s' % subs,
    # [ *3( h16 ":" ) h16 ] "::"    h16 ":"   ls32
    '((%(hex)s:){0,3}%(hex)s)?::%(hex)s:%(ls32)s' % subs,
    # [ *4( h16 ":" ) h16 ] "::"              ls32
    '((%(hex)s:){0,4}%(hex)s)?::%(ls32)s' % subs,
    # [ *5( h16 ":" ) h16 ] "::"              h16
    '((%(hex)s:){0,5}%(hex)s)?::%(hex)s' % subs,
    # [ *6( h16 ":" ) h16 ] "::"
    '((%(hex)s:){0,6}%(hex)s)?::' % subs,
    ]

ipv6 = '(({0})|({1})|({2})|({3})|({4})|({5})|({6})|({7}))'.format(*variations)

ipv_future = 'v[0-9A-Fa-f]+.[%s]+' % (
    important_characters['re_unreserved'] +
    important_characters['re_sub_delimiters'] +
    ':')

ip_literal = '\[({0}|{1})\]'.format(ipv6, ipv_future)

# Pattern for matching the host piece of the authority
HOST_PATTERN = '({0}|{1}|{2})'.format(reg_name, ipv4, ip_literal)

SUBAUTHORITY_MATCHER = re.compile((
    '^(?:(?P<userinfo>[A-Za-z0-9_.~\-%:]+)@)?'  # userinfo
    '(?P<host>{0}?)'  # host
    ':?(?P<port>\d+)?$'  # port
    ).format(HOST_PATTERN))

IPv4_MATCHER = re.compile('^' + ipv4 + '$')


# ####################
# Path Matcher Section
# ####################

# See http://tools.ietf.org/html/rfc3986#section-3.3 for more information
# about the path patterns defined below.

# Percent encoded character values
pct_encoded = '%[A-Fa-f0-9]{2}'
pchar = ('([' + important_characters['re_unreserved']
         + important_characters['re_sub_delimiters']
         + ':@]|%s)' % pct_encoded)
segments = {
    'segment': pchar + '*',
    # Non-zero length segment
    'segment-nz': pchar + '+',
    # Non-zero length segment without ":"
    'segment-nz-nc': pchar.replace(':', '') + '+'
    }

# Path types taken from Section 3.3 (linked above)
path_empty = '^$'
path_rootless = '%(segment-nz)s(/%(segment)s)*' % segments
path_noscheme = '%(segment-nz-nc)s(/%(segment)s)*' % segments
path_absolute = '/(%s)?' % path_rootless
path_abempty = '(/%(segment)s)*' % segments

# Matcher used to validate path components
PATH_MATCHER = re.compile('^(%s|%s|%s|%s|%s)$' % (
    path_abempty, path_absolute, path_noscheme, path_rootless, path_empty
    ))


# ##################################
# Query and Fragment Matcher Section
# ##################################

QUERY_MATCHER = re.compile(
    '^([/?:@' + important_characters['re_unreserved']
    + important_characters['re_sub_delimiters']
    + ']|%s)*$' % pct_encoded)

FRAGMENT_MATCHER = QUERY_MATCHER

# Scheme validation, see: http://tools.ietf.org/html/rfc3986#section-3.1
SCHEME_MATCHER = re.compile('^[A-Za-z][A-Za-z0-9+.\-]*$')

# Relative reference matcher

# See http://tools.ietf.org/html/rfc3986#section-4.2 for details
relative_part = '(//%s%s|%s|%s|%s)' % (
    component_pattern_dict['authority'], path_abempty, path_absolute,
    path_noscheme, path_empty
    )

RELATIVE_REF_MATCHER = re.compile('^%s(\?%s)?(#%s)?$' % (
    relative_part, QUERY_MATCHER.pattern, FRAGMENT_MATCHER.pattern
    ))

# See http://tools.ietf.org/html/rfc3986#section-3 for definition
hier_part = '(//%s%s|%s|%s|%s)' % (
    component_pattern_dict['authority'], path_abempty, path_absolute,
    path_rootless, path_empty
    )

# See http://tools.ietf.org/html/rfc3986#section-4.3
ABSOLUTE_URI_MATCHER = re.compile('^%s:%s(\?%s)?$' % (
    component_pattern_dict['scheme'], hier_part, QUERY_MATCHER.pattern[1:-1]
    ))


# Path merger as defined in http://tools.ietf.org/html/rfc3986#section-5.2.3
def merge_paths(base_uri, relative_path):
    """Merge a base URI's path with a relative URI's path."""
    if base_uri.path is None and base_uri.authority is not None:
        return '/' + relative_path
    else:
        path = base_uri.path or ''
        index = path.rfind('/')
        return path[:index] + '/' + relative_path



# -*- coding: utf-8 -*-
"""
hyper/http20/bufsocket.py
~~~~~~~~~~~~~~~~~~~~~~~~~

This file implements a buffered socket wrapper.

The purpose of this is to avoid the overhead of unnecessary syscalls while
allowing small reads from the network. This represents a potentially massive
performance optimisation at the cost of burning some memory in the userspace
process.
"""
import select
from .exceptions import ConnectionResetError, LineTooLongError


class BufferedSocket(object):
    """
    A buffered socket wrapper.

    The purpose of this is to avoid the overhead of unnecessary syscalls while
    allowing small reads from the network. This represents a potentially
    massive performance optimisation at the cost of burning some memory in the
    userspace process.
    """
    def __init__(self, sck, buffer_size=1000):
        """
        Create the buffered socket.

        :param sck: The socket to wrap.
        :param buffer_size: The size of the backing buffer in bytes. This
            parameter should be set to an appropriate value for your use case.
            Small values of ``buffer_size`` increase the overhead of buffer
            management: large values cause more memory to be used.
        """
        # The wrapped socket.
        self._sck = sck

        # The buffer we're using.
        self._backing_buffer = bytearray(buffer_size)
        self._buffer_view = memoryview(self._backing_buffer)

        # The size of the buffer.
        self._buffer_size = buffer_size

        # The start index in the memory view.
        self._index = 0

        # The number of bytes in the buffer.
        self._bytes_in_buffer = 0

    @property
    def _remaining_capacity(self):
        """
        The maximum number of bytes the buffer could still contain.
        """
        return self._buffer_size - self._index

    @property
    def _buffer_end(self):
        """
        The index of the first free byte in the buffer.
        """
        return self._index + self._bytes_in_buffer

    @property
    def can_read(self):
        """
        Whether or not there is more data to read from the socket.
        """
        read = select.select([self._sck], [], [], 0)[0]
        if read:
            return True

        return False

    @property
    def buffer(self):
        """
        Get access to the buffer itself.
        """
        return self._buffer_view[self._index:self._buffer_end]

    def advance_buffer(self, count):
        """
        Advances the buffer by the amount of data consumed outside the socket.
        """
        self._index += count
        self._bytes_in_buffer -= count

    def new_buffer(self):
        """
        This method moves all the data in the backing buffer to the start of
        a new, fresh buffer. This gives the ability to read much more data.
        """
        def read_all_from_buffer():
            end = self._index + self._bytes_in_buffer
            return self._buffer_view[self._index:end]

        new_buffer = bytearray(self._buffer_size)
        new_buffer_view = memoryview(new_buffer)
        new_buffer_view[0:self._bytes_in_buffer] = read_all_from_buffer()

        self._index = 0
        self._backing_buffer = new_buffer
        self._buffer_view = new_buffer_view

        return

    def recv(self, amt):
        """
        Read some data from the socket.

        :param amt: The amount of data to read.
        :returns: A ``memoryview`` object containing the appropriate number of
            bytes. The data *must* be copied out by the caller before the next
            call to this function.
        """
        # In this implementation you can never read more than the number of
        # bytes in the buffer.
        if amt > self._buffer_size:
            amt = self._buffer_size

        # If the amount of data we've been asked to read is less than the
        # remaining space in the buffer, we need to clear out the buffer and
        # start over.
        if amt > self._remaining_capacity:
            self.new_buffer()

        # If there's still some room in the buffer, opportunistically attempt
        # to read into it.
        # If we don't actually _need_ the data (i.e. there's enough in the
        # buffer to satisfy the request), use select to work out if the read
        # attempt will block. If it will, don't bother reading. If we need the
        # data, always do the read.
        if self._bytes_in_buffer >= amt:
            should_read = select.select([self._sck], [], [], 0)[0]
        else:
            should_read = True

        if (self._remaining_capacity > self._bytes_in_buffer and should_read):
            count = self._sck.recv_into(self._buffer_view[self._buffer_end:])

            # The socket just got closed. We should throw an exception if we
            # were asked for more data than we can return.
            if not count and amt > self._bytes_in_buffer:
                raise ConnectionResetError()
            self._bytes_in_buffer += count

        # Read out the bytes and update the index.
        amt = min(amt, self._bytes_in_buffer)
        data = self._buffer_view[self._index:self._index+amt]

        self._index += amt
        self._bytes_in_buffer -= amt

        return data

    def fill(self):
        """
        Attempts to fill the buffer as much as possible. It will block for at
        most the time required to have *one* ``recv_into`` call return.
        """
        if not self._remaining_capacity:
            self.new_buffer()

        count = self._sck.recv_into(self._buffer_view[self._buffer_end:])
        if not count:
            raise ConnectionResetError()

        self._bytes_in_buffer += count

        return

    def readline(self):
        """
        Read up to a newline from the network and returns it. The implicit
        maximum line length is the buffer size of the buffered socket.

        Note that, unlike recv, this method absolutely *does* block until it
        can read the line.

        :returns: A ``memoryview`` object containing the appropriate number of
            bytes. The data *must* be copied out by the caller before the next
            call to this function.
        """
        # First, check if there's anything in the buffer. This is one of those
        # rare circumstances where this will work correctly on all platforms.
        index = self._backing_buffer.find(
            b'\n',
            self._index,
            self._index + self._bytes_in_buffer
        )

        if index != -1:
            length = index + 1 - self._index
            data = self._buffer_view[self._index:self._index+length]
            self._index += length
            self._bytes_in_buffer -= length
            return data

        # In this case, we didn't find a newline in the buffer. To fix that,
        # read some data into the buffer. To do our best to satisfy the read,
        # we should shunt the data down in the buffer so that it's right at
        # the start. We don't bother if we're already at the start of the
        # buffer.
        if self._index != 0:
            self.new_buffer()

        while self._bytes_in_buffer < self._buffer_size:
            count = self._sck.recv_into(self._buffer_view[self._buffer_end:])
            if not count:
                raise ConnectionResetError()

            # We have some more data. Again, look for a newline in that gap.
            first_new_byte = self._buffer_end
            self._bytes_in_buffer += count
            index = self._backing_buffer.find(
                b'\n',
                first_new_byte,
                first_new_byte + count,
            )

            if index != -1:
                # The length of the buffer is the index into the
                # buffer at which we found the newline plus 1, minus the start
                # index of the buffer, which really should be zero.
                assert not self._index
                length = index + 1
                data = self._buffer_view[:length]
                self._index += length
                self._bytes_in_buffer -= length
                return data

        # If we got here, it means we filled the buffer without ever getting
        # a newline. Time to throw an exception.
        raise LineTooLongError()

    def __getattr__(self, name):
        return getattr(self._sck, name)



from .protocols.http import request as http_request, request_handler as http_request_handler
from .protocols.ws import request_handler as ws_request_handler, request as ws_request
from .protocols.grpc import request as grpc_request
from .dbapi import request as dbo_request
from .protocols.proxy import tunnel_handler
from .protocols.http import localstorage as ls, util

def make_ws (_method, url, params, auth, headers, meta, proxy, logger):
    req = ws_request.Request (url, params, headers, auth, logger, meta)
    if proxy:
        if _method == 'wss':
            handler_class = tunnel_handler.WSSSLProxyTunnelHandler
        else:
            handler_class = tunnel_handler.WSProxyTunnelHandler
    else:        
        handler_class = ws_request_handler.RequestHandler
    req.handler = handler_class    
    return req
    
def make_http (_method, url, params, auth, headers, meta, proxy, logger):
    headers = util.normheader (headers)
    if ls.g:
        headers ['Cookie'] = ls.g.get_cookie_as_string (url)
    if proxy and url.startswith ('https://'):
        handler_class = tunnel_handler.SSLProxyTunnelHandler
    else:            
        handler_class = http_request_handler.RequestHandler

    if _method == "rpc":
        rpcmethod, params = params
        req = http_request.XMLRPCRequest (url, rpcmethod, params, headers, auth, logger, meta)
    
    elif _method == "jsonrpc":
        rpcmethod, params = params
        req = http_request.JSONRPCRequest (url, rpcmethod, params, headers, auth, logger, meta)    

    elif _method == "grpc":
        rpcmethod, params = params
        req = grpc_request.GRPCRequest (url, rpcmethod, params, headers, auth, logger, meta)
    
    elif _method == "upload":
        req = http_request.HTTPMultipartRequest (url, "POST", params, headers, auth, logger, meta)
                
    else:
        ct, ac = "application/x-www-form-urlencoded", "*/*"
        if _method.endswith ("json"):
            _method = _method [:-4]
            ct, ac = "application/json", "application/json"
        util.set_content_types (headers, params, (ct, ac))
        req = http_request.HTTPRequest (url, _method.upper (), params, headers, auth, logger, meta)        
    
    req.handler = handler_class
    return req

def make_dbo (_method, server, dbmethod, params, dbname, auth, meta, logger):
    return dbo_request.Request (_method [1:], server, dbname, auth, dbmethod, params, None, meta)



# -*- coding: utf-8 -*-
"""
hyper/http20/exceptions
~~~~~~~~~~~~~~~~~~~~~~~

This defines exceptions used in the HTTP/2 portion of hyper.
"""


class HTTP20Error(Exception):
    """
    The base class for all of ``hyper``'s HTTP/2-related exceptions.
    """
    pass


class HPACKEncodingError(HTTP20Error):
    """
    An error has been encountered while performing HPACK encoding.
    """
    pass


class HPACKDecodingError(HTTP20Error):
    """
    An error has been encountered while performing HPACK decoding.
    """
    pass


class ConnectionError(HTTP20Error):
    """
    The remote party signalled an error affecting the entire HTTP/2
    connection, and the connection has been closed.
    """
    pass


class ProtocolError(HTTP20Error):
    """
    The remote party violated the HTTP/2 protocol.
    """
    pass


class StreamResetError(HTTP20Error):
    """
    A stream was forcefully reset by the remote party.
    """
    pass

class StreamClosedError(HTTP20Error):
    """
    A stream was forcefully reset by the remote party.
    """
    pass



import time
import threading

DEFAULT_ZOMBIE_TIMEOUT = 60
DEFAULT_KEEP_ALIVE = 2

class ConnectProxy:
    def __init__ (self, asyncon):
        self._asyncon = asyncon
        self._args = None
        self._timeout = 30
        self._cv = threading.Condition ()

    def __getattr__ (self, attr):
        return getattr (self._asyncon, attr)

    def set_timeout (self, timeout):
        self._timeout = timeout

    def set_active (self, flag):
        if flag:
            return
        # exception occured
        self._args = None

    def _canceled (self):
        # must be called from very first line of execute_late
        if self._args is None:
            self._asyncon.set_active (False)
            return True
        return False

    def execute (self, *args):
        self._args = args
        self._asyncon.add_task (self.execute_late)

    def execute_late (self):
        if self._canceled ():
            return
        raise NotImplementedError


class BaseConnect:
    def __init__ (self, lock, logger):
        self.lock = lock
        self.logger = logger

        self.active = 0
        self.request_count = 0
        self.event_time = 0
        self.backend = False
        self.zombie_timeout = DEFAULT_ZOMBIE_TIMEOUT
        self.keep_alive = DEFAULT_KEEP_ALIVE

        self._history = []
        self._tasks = []
        self._no_more_request = False
        self._cv = threading.Condition ()

    def get_history (self):
        return self._history

    def log_history (self, msg):
        self._history.append (msg)

    def set_backend (self, backend_keep_alive = 10):
        self.backend = True
        self.keep_alive = backend_keep_alive

    def get_request_count (self):
        return self.request_count

    def set_timeout (self, timeout = 10):
        # CAUTION: used at proxy.tunnel_handler
        self.zombie_timeout = timeout

    def set_event_time (self):
        self.event_time = time.time ()

    def _set_active (self, flag):
        if flag:
            flag = time.time ()
            self.request_count += 1
        else:
            flag = 0
            self.set_timeout (self.keep_alive)
            if self._tasks:
                self._set_active (True)
                return self._tasks.pop (0)
        self.active = flag

    def set_active (self, flag, nolock = False):
        if nolock or self.lock is None:
            task = self._set_active (flag)
        else:
            with self.lock:
                task = self._set_active (flag)
        task and task ()

    def _add_task (self, task):
        if self.active > 0:
            self._tasks.append (task)
            return
        self._set_active (True)
        return task

    def add_task (self, task):
        if self.lock is None:
            task = self._add_task (task)
        else:
            with self.lock:
                task = self._add_task (task)
        task and task ()

    def get_active (self, nolock = False):
        if nolock or self.lock is None:
            return self.active
        with self.lock:
            return self.active

    def isactive (self):
        return self.get_active () > 0



from hashlib import md5
from urllib.parse import urlparse
from . import util
import re

DEFAULT_PORT_MAP = {'http': 80, 'https': 443, 'ws': 80, 'wss': 443}
NON_IP_COMPONENT = re.compile ("[^.0-9]")

def sort_args (data):
	args = {}
	for arg in data.split ("&"):
		try: k, v = arg.split ("=")
		except: continue
		if v:
			args [k] = v
	args = list(args.items ())
	args.sort ()
	argslist = []
	for k, v in args:
		argslist.append ("%s=%s" % (k, v))		
	argslist.sort ()		
	return "&".join (argslist)

def parse_address (scheme, address):
	try: 
		host, port = address.split (":", 1)
		port = int (port)
	except ValueError:
		host = address
		if scheme in ("http", "ws"):
			port = 80
		else:
			port = 443	
	return host, port
	
def uuid (uri, method = "get", data = "", include_data = True):
	# url id
	if uri.find ("://") == -1:
		return None
	method = method.lower ()
	scheme, address, script, params, qs, fragment = urlparse (uri)
	
	if not script: 
		script = "/"
	host, port = parse_address (scheme, address)		
	sig = [
		method,
		host.startswith ("www.") and host [4:] or host, 
		str (port),
		script
	]
	if include_data:
		sig.append (params)
	if qs:
		sig.append (sort_args (qs))
	if data:
		sig.append (sort_args (data))
	return md5 (":".join (sig).encode ("utf8")).hexdigest ()

def usid (uri, method = "get"):	
	# script id
	return make_uuid (uri, method, "", False)

def uinf (url):
	return UrlInfo (url)

	
class UrlInfo:	
	def __init__ (self, url):
		self.url = url
		self._parse ()
		
	def _parse (self):
		url = self.url
		self.scheme, self.netloc, self.script, self.params, self.querystring, self.fragment = urlparse (url)
		
		if not self.script: 
			self.script = '/'
		if self.querystring:
			try: self.querystring = util.strencode (self.querystring)
			except: pass
		
		uri = self.script #real request uri
		if self.params:
			uri += ';' + self.params
		if self.querystring:
			uri += '?' + self.querystring
		self.uri = uri
		
		position = self.netloc.find('@')
		if position > -1:
			self.auth, self.netloc = self.netloc.split ("@", 1)		
			try: 
				self.username, self.password = self.auth.split (':', 1)
			except ValueError:	
				pass
			else:	
				self.auth = (self.username, self.password)

		try:
			self.netloc, self.port = self.netloc.split (':', 1)
		except ValueError:
			try:
				self.netloc, self.port = self.netloc, DEFAULT_PORT_MAP [self.scheme]
			except KeyError:
				self.netloc, self.port = self.netloc, 80
		
		self.netloc = self.netloc.lower ()
		try: self.port = int (self.port)
		except: self.port = 80
		
		if DEFAULT_PORT_MAP [self.scheme] == self.port:
			self.rfc = '%s://%s%s' % (self.scheme, self.netloc, self.uri)
		else:
			self.rfc = '%s://%s:%d%s' % (self.scheme, self.netloc, self.port, self.uri)
			
		self.host = None
		self.domain = None
		self.subdomain = None
		self.topdomain = None
		
		if NON_IP_COMPONENT.search (self.netloc):
			netloc = self.netloc
			netloc2 = netloc.split (".")
			self.topdomain = netloc2 [-1]
			
			if len (netloc2) == 1:
				self.host = ""	
				self.domain = netloc
				self.subdomain = netloc				
			elif len (netloc2) == 2:
				self.host = ""	
				self.domain = netloc
				self.subdomain = netloc2 [0]				
			elif len (netloc2 [-1]) >= 3:
				self.domain = ".".join (netloc2 [-2:])
				self.host = ".".join (netloc2 [:-2])
				self.subdomain = netloc2 [-2]			
				if not self.host:
					self.host = "www"
			elif len (netloc2 [-1]) == 2:
				if len (netloc2) == 2:
					self.domain = ".".join (netloc2)
					self.subdomain = netloc2 [0]
					self.host = "www"
				elif len (netloc2) == 3:
					if netloc2 [0] == "www":
						self.domain = ".".join (netloc2 [-2:])
						self.subdomain = netloc2 [-2]
					else:
						self.domain = ".".join (netloc2)
						self.subdomain = netloc2 [0]
					self.host = "www"
				else:
					self.domain = ".".join (netloc2 [-3:])
					self.host = ".".join (netloc2 [:-3])
					self.subdomain = netloc2 [-3]
		